package net.wanji.utc.service.impl;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.hikvision.artemis.sdk.ArtemisHttpUtil;
import com.hikvision.artemis.sdk.config.ArtemisConfig;
import lombok.extern.slf4j.Slf4j;
import net.wanji.utc.common.Result;
import net.wanji.utc.common.baseentity.BaseCrossInfo;
import net.wanji.utc.common.exception.ControlException;
import net.wanji.utc.common.typeenum.BasicEnum;
import net.wanji.utc.mapper.CrossPhaseMapper;
import net.wanji.utc.po.CrossPhasePO;
import net.wanji.utc.service.ControlCommandService;
import net.wanji.utc.service.HkGetSignalMethodService;
import net.wanji.utc.vo.PhaseLockVO;
import net.wanji.utc.vo.plansend.PlanSendVO;
import net.wanji.utc.vo.plansend.SectionPlan;
import net.wanji.utc.vo.plansend.TimeSlice;
import net.wanji.utc.vo.signal.SignalLightStateVo;
import net.wanji.utc.vo.signal.SignalRingVo;
import net.wanji.utc.vo.timeplan.Phase;
import net.wanji.utc.vo.timeplan.Ring;
import net.wanji.utc.vo.timeplan.TimePlanVO;
import org.springframework.stereotype.Service;

import javax.annotation.Resource;
import java.util.*;
import java.util.stream.Collectors;

import static net.wanji.utc.common.constant.Constants.*;

/**
 * @author wuxiaokai
 * @date 2022/11/21 9:51:47
 */
@Slf4j
@Service("hkControlCommandService")
public class HKControlCommandServiceImpl implements ControlCommandService {

    @Resource
    private HkGetSignalMethodService hkGetSignalMethodService;

    @Resource
    private ArtemisConfig artemisConfig;

    @Resource
    private CrossPhaseMapper crossPhaseMapper;

    @Override
    public <T> Result<T> schemeSend() {
        return null;
    }

    @Override
    public <T> Result<T> planSend(PlanSendVO planSendVO) throws Exception {
        if (sendPlanInfo(planSendVO)) {
            if (sendTimeInfo(planSendVO)) {
                return Result.success();
            }
        }
        return Result.error("下发指令失败");
    }

    @Override
    public <T> Result<T> scheduleSend(String signalId, String sourceType, String signalType, Integer command) {
        return controlSignal(signalId, NORMAL_RUNNING_CONTROL);
    }

    @Override
    public <T> Result<T> lockControl(String signalId, PhaseLockVO phaseLock) throws Exception {
        if (phaseLock.getCommand().equals(TRUE)) {
            return updateManualLock(signalId, phaseLock.getLockTime());
        } else {
            return controlSignal(signalId, UNLOCK_RUNNING_CONTROL);
        }
    }

    private <T> Result<T> updateManualLock(String signalId, Integer lockTime) throws Exception {
        JSONObject param = new JSONObject();
        param.put("crossCode", signalId);
        param.put("controlType", LOCK_RUNNING_CONTROL);
        setSignalControlLaneInfos(signalId, param);
        param.put("controlNo", 0);
        //控制时长，单位 s，0 则持续控制
        param.put("duration", lockTime * 60);
        return setSignalControl(param);
    }

    /**
     * 控制命令为灯态锁定时，设置相关车道参数信息
     *
     * @param signalId, 信号机id
     */
    private void setSignalControlLaneInfos(String signalId, JSONObject obj) throws Exception {
        //获取当前运行方案的相位id
        List<SignalLightStateVo> signalInfos = getSignalInfoVos(signalId);
        List<SignalRingVo> runrings = signalInfos.get(0).getRings();
        //HK 为单环
        Integer phaseId = Integer.valueOf(runrings.get(0).getPhaseId());
        //获取某一路口下海康相位和车道关系
        Map<Integer, JSONArray> laneArrayMap = new HashMap<>();
        getPhaseRelationLanes(signalId, laneArrayMap, new HashMap<>());
        obj.put("laneInfos", laneArrayMap.get(phaseId));
    }

    /**
     * 获取某一路口下海康相位和车道关系
     *
     * @param crossCode    信号机唯一编号
     * @param laneArrayMap 相位与车道集合
     * @param laneNoMap    转向和方向与车道关系集合
     */
    private void getPhaseRelationLanes(String crossCode, Map<Integer, JSONArray> laneArrayMap, Map<String, Integer> laneNoMap) throws Exception {
        Map<String, String> path = getPathMapByApiCode("queryPhaseRlatLanes");
        Map<String, String> queryMap = new HashMap<>();
        queryMap.put("crossCode", crossCode);
        String strResult = ArtemisHttpUtil.doPostStringArtemis(artemisConfig, path, null, queryMap, null, "application/json", null);
        JSONObject object = JSON.parseObject(strResult);
        if (HK_SUCCESS_CODE.equals(object.getInteger(HK_CODE_KEY))) {
            JSONObject data = object.getJSONObject("data");
            JSONArray phaseList = data.getJSONArray("phaseList");
            for (int i = 0; i < phaseList.size(); i++) {
                JSONObject lanesObject = phaseList.getJSONObject(i);
                Integer phaseNo = lanesObject.getInteger("phaseNo");
                JSONArray laneList = lanesObject.getJSONArray("lanes");
                // 存入相位与车道集合
                laneArrayMap.put(phaseNo, laneList);
                for (int j = 0; j < laneList.size(); j++) {
                    JSONObject laneObject = laneList.getJSONObject(j);
                    String key = laneObject.getString("direction") + SEPARATOR_UNDER_LINE + laneObject.getString("turn");
                    // 存入方向和转向对应的车道编号
                    laneNoMap.put(key, laneObject.getInteger("laneNo"));
                }
            }
        }
    }

    @Override
    public <T> Result<T> stepControl(String signalId, String sourceType, String signalType, Integer command, Integer stepNum) {
        if (command.equals(TRUE)) {
            return updateManualStep(signalId, stepNum);
        } else {
            return controlSignal(signalId, CANCEL_STEP_CONTROL);
        }
    }

    @Override
    public <T> Result<T> allRedControl(String signalId, String sourceType, String signalType, Integer command) {
        Integer runningMode = command.equals(TRUE) ? ALL_RED_CONTROL : NORMAL_RUNNING_CONTROL;
        return controlSignal(signalId, runningMode);
    }

    @Override
    public <T> Result<T> yellowLightControl(String signalId, String sourceType, String signalType, Integer command) {
        Integer runningMode = command.equals(TRUE) ? YELLOW_LIGHT_CONTROL : NORMAL_RUNNING_CONTROL;
        return controlSignal(signalId, runningMode);
    }

    @Override
    public <T> Result<T> closeLightControl(String signalId, String sourceType, String signalType, Integer command) {
        Integer runningMode = command.equals(TRUE) ? CLOSE_LIGHT_CONTROL : NORMAL_RUNNING_CONTROL;
        return controlSignal(signalId, runningMode);
    }

    @Override
    public <T> Result<T> recoverSchedule(String signalId, String sourceType, String signalType, Integer command) {
        return null;
    }

    @Override
    public <T> Result<T> phaseTimingSend(List<TimePlanVO> timePlans) {
        try {
            for (TimePlanVO timePlanVo : timePlans) {
                if (!phaseTimingSend(timePlanVo)) {
                    //发送失败
                    log.error("批量下发配时方案信息到路口信号机或信号系统中异常！");
                    throw new ControlException("批量下发配时方案信息到路口信号机或信号系统中异常！");
                }
                //暂停1秒
                Thread.sleep(1000);
            }
            return Result.success();
        } catch (InterruptedException e) {
            log.error("批量配时方案下发-->线程等待异常！{}", e.getMessage());
            throw new ControlException("批量配时方案下发-->线程等待异常！");
        }
    }

    private boolean phaseTimingSend(TimePlanVO timePlan) {
        try {
            //获取某一路口下海康相位和车道关系
            Map<Integer, JSONArray> laneArrayMap = new HashMap<>();
            getPhaseRelationLanes(timePlan.getCode(), laneArrayMap, new HashMap<>());
            //拼接参数
            JSONObject bodyObjectParam = new JSONObject();
            JSONArray patternArrayParam = new JSONArray();
            JSONObject patternObjectParam = new JSONObject();
            patternObjectParam.put("patternNo", timePlan.getTimePlanId());
            patternObjectParam.put("offSet", timePlan.getOffset());
            patternObjectParam.put("cycle", timePlan.getCycleLen());
            List<Ring> ringList = timePlan.getRings();
            JSONArray ringsArrayParam = new JSONArray();
            for (Ring ring : ringList) {
                List<Phase> phaseList = ring.getPhases();
                JSONArray phaseArrayParam = new JSONArray();
                for (Phase phase : phaseList) {
                    JSONObject phaseObjectParam = new JSONObject();
                    phaseObjectParam.put("phaseNo", phase.getPhaseId());
                    phaseObjectParam.put("phaseName", phase.getPhaseName());
                    phaseObjectParam.put("phaseLength", phase.getPhaseTime());
                    phaseObjectParam.put("phaseSequence", phase.getPhaseOrderId());
                    //车道列表参数，适配海康车道参数
                    if (laneArrayMap.containsKey(Integer.valueOf(phase.getPhaseId()))) {
                        phaseObjectParam.put("lanes", laneArrayMap.get(Integer.valueOf(phase.getPhaseId())));
                    }
                    phaseArrayParam.add(phaseObjectParam);
                }
                JSONObject ringObjectParam = new JSONObject();
                ringObjectParam.put("ringNo", ring.getRingNo());
                ringObjectParam.put("phaseList", phaseArrayParam);
                ringsArrayParam.add(ringObjectParam);
            }
            patternObjectParam.put("rings", ringsArrayParam);
            patternArrayParam.add(patternObjectParam);
            bodyObjectParam.put("crossCode", timePlan.getCode());
            bodyObjectParam.put("patternList", patternArrayParam);
            //获取下发配时方案的url
            String strResult = ArtemisHttpUtil.doPostStringArtemis(artemisConfig, getPathMapByApiCode("updateTimePlans"), bodyObjectParam.toJSONString(), null, null, "application/json", null);
            JSONObject object = JSON.parseObject(strResult);
            return HK_SUCCESS_CODE.equals(object.getInteger(HK_CODE_KEY));
        } catch (Exception e) {
            log.error("系统繁忙，服务器端内部错误！", e);
            throw new ControlException("系统繁忙，服务器端内部错误！");
        }
    }

    @Override
    public <T> Result<T> phaseDiffSend() {
        return null;
    }

    private <T> Result<T> updateManualStep(String signalId, Integer stepNum) {
        JSONObject bodyObjectParam = new JSONObject();
        bodyObjectParam.put("crossCode", signalId);
        bodyObjectParam.put("controlType", STEP_CONTROL);
        int controlNo = 0;
        if (0 != stepNum) {
            //根据当前实时灯态获取运行的相位
            SignalLightStateVo signalInfoVo = getSignalInfoVos(signalId).get(0);
            //暂时按单环处理
            SignalRingVo signalRunring = signalInfoVo.getRings().get(0);
            String phaseId = signalRunring.getPhaseId();
            //获取方案数据列表
            List<CrossPhasePO> baseCrossPhasePlanList = crossPhaseMapper.selectByCrossIdAndPlanId(signalInfoVo.getCode(), signalInfoVo.getPhasePlanId());
            Map<String, List<CrossPhasePO>> groupByPhase = baseCrossPhasePlanList.stream().collect(Collectors.groupingBy(CrossPhasePO::getPhaseNo));
            //根据相位获取当前相序
            int newPhaseOrderId = 0;
            for (String key : groupByPhase.keySet()) {
                CrossPhasePO crossPhase = groupByPhase.get(key).get(0);
                if (crossPhase.getPhaseNo().equals(phaseId)) {
                    newPhaseOrderId = (crossPhase.getSort() + stepNum) % groupByPhase.size();
                }
            }
            //根据相序获取相位
            for (String key : groupByPhase.keySet()) {
                CrossPhasePO crossPhase = groupByPhase.get(key).get(0);
                if (newPhaseOrderId == crossPhase.getSort()) {
                    controlNo = Integer.parseInt(crossPhase.getPhaseNo());
                }
            }
        }
        bodyObjectParam.put("controlNo", controlNo);
        return setSignalControl(bodyObjectParam);
    }

    /**
     * @param signalId    信号机编号
     * @param runningMode 0-多时段控制模式 1-关灯、2-黄闪、3-全红、11-取消步进控制、12-灯态锁定、13-灯态解锁
     */
    private <T> Result<T> controlSignal(String signalId, Integer runningMode) {
        //执行恢复时间表方案命令
        if (NORMAL_RUNNING_CONTROL.equals(runningMode) || CANCEL_STEP_CONTROL.equals(runningMode)) {
            //根据实时灯态获取当前运行模式
            SignalLightStateVo signalInfoVo = getSignalInfoVos(signalId).get(0);
            //如果当前模式已经是定周期，就不在执行恢复时间表方案指令
            if (FIXED_CYCLE_MODE.equals(Integer.valueOf(signalInfoVo.getRunMode()))) {
                throw new ControlException("当前控制模式是定周期控制，请勿下发恢复命令");
            }
        }

        JSONObject bodyObjectParam = new JSONObject();
        bodyObjectParam.put("crossCode", signalId);
        bodyObjectParam.put("controlType", runningMode);
        bodyObjectParam.put("controlNo", 0);
        //控制时长，单位 s，0 则持续控制
        bodyObjectParam.put("duration", 0);
        return setSignalControl(bodyObjectParam);
    }

    private <T> Result<T> setSignalControl(JSONObject bodyObjectParam) {
        try {
            String strResult = ArtemisHttpUtil.doPostStringArtemis(artemisConfig, getPathMapByApiCode("updateLockPhase"), bodyObjectParam.toJSONString(), null, null, "application/json", null);
            JSONObject object = JSON.parseObject(strResult);
            if (HK_SUCCESS_CODE.equals(object.getInteger(HK_CODE_KEY))) {
                return Result.success();
            }
        } catch (Exception e) {
            log.error("系统繁忙，服务器端内部错误！", e);
        }
        return Result.error("下发指令失败");
    }

    /**
     * 根据信号机获取灯态列表
     */
    private List<SignalLightStateVo> getSignalInfoVos(String signalId) {
        List<BaseCrossInfo> baseSignals = new ArrayList<>();
        BaseCrossInfo baseSignal = new BaseCrossInfo();
        baseSignal.setCode(signalId);
        baseSignal.setManufacturerAbbr(BasicEnum.ManufacturerEnum.HK.getAbbr());
        baseSignals.add(baseSignal);
        return hkGetSignalMethodService.queryHkSignalInfo(baseSignals);
    }

    /**
     * 下发时段信息
     *
     * @param planSendVO 计划下发VO
     * @return boolean
     */
    private boolean sendTimeInfo(PlanSendVO planSendVO) throws Exception {
        //拼接参数
        JSONObject bodyObjectParam = new JSONObject();
        // 拼接参数 海康后续修改不传入 controlType ，controlType 取值默认为原始值
        List<SectionPlan> sectionPlans = planSendVO.getSectionPlans();
        for (SectionPlan plan : sectionPlans) {
            //拼接参数，信号机方案对象
            JSONObject signalObjectParam = new JSONObject();
            //拼接参数，信号列表
            JSONArray signalArrayParam = new JSONArray();
            //拼接参数，方案列表
            JSONArray planArrayParam = new JSONArray();
            //拼接参数，时段列表
            JSONArray sectionArrayParam = new JSONArray();
            List<TimeSlice> timeSlices = plan.getTimeSlices();
            for (TimeSlice timeSlice : timeSlices) {
                JSONObject section = new JSONObject();
                section.put("timeSecNo", timeSlice.getTimeSliceId());
                section.put("patternNo", timeSlice.getTimePlanId());
                section.put("beginTime", timeSlice.getStartTime());
                sectionArrayParam.add(section);
            }
            JSONObject planObjectParam = new JSONObject();
            planObjectParam.put("planNo", plan.getSectionPlanId());
            planObjectParam.put("planName", plan.getSectionPlanDesc());
            planObjectParam.put("sectionList", sectionArrayParam);
            planArrayParam.add(planObjectParam);
            signalObjectParam.put("crossCode", planSendVO.getCode());
            signalObjectParam.put("planList", planArrayParam);
            signalArrayParam.add(signalObjectParam);
            bodyObjectParam.put("data", signalArrayParam);
        }
        // 下发时段方案
        String strResult = ArtemisHttpUtil.doPostStringArtemis(artemisConfig, getPathMapByApiCode("updateSectionPlans"), bodyObjectParam.toJSONString(), null, null, "application/json", null);
        JSONObject object = JSON.parseObject(strResult);
        return object.getInteger(HK_CODE_KEY).equals(HK_SUCCESS_CODE);
    }

    /**
     * 下发计划信息
     *
     * @param planSendVO 计划下发VO
     * @return boolean
     * @throws Exception 异常
     */
    private boolean sendPlanInfo(PlanSendVO planSendVO) throws Exception {
        JSONArray body = new JSONArray();
        body.add(planSendVO.getCode());
        Map<String, String> path = getPathMapByApiCode("queryRunPlan");
        String strResult = ArtemisHttpUtil.doPostStringArtemis(artemisConfig, path, body.toString(), null, null, "application/json", null);
        JSONObject jsonObj = JSON.parseObject(strResult);
        //日期 id
        int id = 1;
        if (jsonObj.getInteger(HK_CODE_KEY).equals(HK_SUCCESS_CODE)) {
            // 获取日期列表
            JSONArray schedulesArray = jsonObj.getJSONArray("data").getJSONObject(0).getJSONArray("schedules");
            JSONObject scheduleObj = schedulesArray.getJSONObject(schedulesArray.size() - 1);
            id = scheduleObj.getInteger("id");
        }
        //拼接参数
        JSONObject bodyObjectParam = new JSONObject();
        JSONArray schedulesArrayParam = new JSONArray();
        // 获取传入的星期数组
        JSONArray weeks = JSON.parseArray(JSON.toJSONString(planSendVO.getWeeks()));
        // 获取传入的特殊日期数组
        JSONArray specialDays = JSON.parseArray(JSON.toJSONString(planSendVO.getSpecialDays()));
        if (null != weeks) {
            //去重，获取所有的运行方案号
            Set<String> set = new HashSet<>();
            for (int i = 0; i < weeks.size(); i++) {
                String sectionPlanId = weeks.getJSONObject(i).getString("sectionPlanId");
                set.add(sectionPlanId);
            }
            if (null != specialDays) {
                for (int i = 0; i < specialDays.size(); i++) {
                    String sectionPlanId = specialDays.getJSONObject(i).getString("sectionPlanId");
                    set.add(sectionPlanId);
                }
            }
            int num = 0;
            for (String planNo : set) {
                JSONObject schedulesObjectParam = new JSONObject();
                List<Integer> weeksListParam = new ArrayList<>();
                // 遍历星期列表，拼接星期集合参数
                for (int i = 0; i < weeks.size(); i++) {
                    JSONObject week = weeks.getJSONObject(i);
                    String weekNum = week.getString("weekNum");
                    String sectionPlanId = week.getString("sectionPlanId");
                    if (planNo.equals(sectionPlanId)) {
                        weeksListParam.add(Integer.valueOf(weekNum));
                    }
                }
                // 遍历特殊日期列表，拼接特殊日期集合参数
                if (null != specialDays) {
                    List<String> specialDaysListParam = new ArrayList<>();
                    for (int i = 0; i < specialDays.size(); i++) {
                        JSONObject specialDay = specialDays.getJSONObject(i);
                        String dateStr = specialDay.getString("dateStr");
                        String sectionPlanId = specialDay.getString("sectionPlanId");
                        if (planNo.equals(sectionPlanId)) {
                            specialDaysListParam.add(dateStr);
                        }
                    }
                    schedulesObjectParam.put("dates", specialDaysListParam);
                    //0 说明 dates 内的日期是日期集合，1 说明是dates 内的数据是日期范围
                    schedulesObjectParam.put("isPeriod", 0);
                }
                //scheduleNo 日期号 ，支持返回【1~128】
                if (id + set.size() > 128) {
                    schedulesObjectParam.put("scheduleNo", 128 - set.size() - (++num));
                } else {
                    schedulesObjectParam.put("scheduleNo", id + (++num));
                }
                schedulesObjectParam.put("planNo", planNo);
                schedulesObjectParam.put("weeks", weeksListParam);
                schedulesArrayParam.add(schedulesObjectParam);
            }
        }
        bodyObjectParam.put("crossCode", planSendVO.getCode());
        bodyObjectParam.put("schedules", schedulesArrayParam);
        // 下发路口运行计划方案
        String updateResult = ArtemisHttpUtil.doPostStringArtemis(artemisConfig, getPathMapByApiCode("updateRunPlan"), bodyObjectParam.toJSONString(), null, null, "application/json", null);
        JSONObject updateObject = JSON.parseObject(updateResult);
        return updateObject.getInteger(HK_CODE_KEY).equals(HK_SUCCESS_CODE);
    }
}
